/*
   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
   3. Neither the name of the project nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

//#define DEBUG
#ifdef DEBUG
#define DebugPrintf0( message ) printf( message ); fflush( stdout )
#define DebugPrintf1( message , a ) printf( message , (a) ); fflush( stdout )
#define DebugPrintf2( message , a, b ) printf( message , (a), (b) ); fflush( stdout )
#else
#define DebugPrintf0( message )
#define DebugPrintf1( message , a )
#define DebugPrintf2( message , a, b )
#endif

#include "AMG.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "myCUDA.h"
#include "mySMOOTHER.h"
#include "myTime.h"
//#include "bmp.h"

#define PADDING_BYTE 256
#define MAT(a, col, row, x, y) (a)[(y)*(row) + (x)]

void LU_Decomposition(lu_crs *lu_a, double *a_l, double *a_u);
	void UnpackLU_CRS(lu_crs *lu_a, double *a);

int SetColor(lu_crs *a, int *color, double threshold, int num);
	void MaxRowVal(lu_crs *a, double *max);
	bool PaintColor(lu_crs *a, int pallet, int *color, double *max, double threshold);

void SetHistogram(int m, int n, int *histogram, int *data);
void PrintHistogram(int n, int *histogram);

int *SetOrdering(color_object *cl_obj, int np);

void PrintBMP(lu_crs *a, int *ordering, int rank);

void SendLU_CRS(lu_crs *a, lu_icrss &da, color_object *cl_obj, int *ordering_col, int *ordering_row);
void SendCRS(crs *a, icrss &da, color_object *cl_obj, int *ordering_col, int *ordering_row);

void LU(level_object *lv_obj);
	void ForwardElimination(double *l, double *b, double *x, int np);
	void BackwardSubstitution(double *u, double *b, double *x, int np);

void AMG_SETUP(object *obj, int lv)
{
	DebugPrintf0("AMG SETUP Start.\n");

	int		i, np, size;
	int		*ordering_col, *ordering_row;
	double		*l, *u;
	level_object	*lv_obj;
	color_object	*cl_obj;

	lv_obj = &obj->level[lv];
	cl_obj = &lv_obj->cl_obj;

	/* ɤʬ */
	cl_obj->length = lv_obj->a.n;
	cl_obj->color = (int *)CUDA_MALLOC_HOST(sizeof(int) * lv_obj->a.n, "color");
	cl_obj->colorNum = SetColor(&lv_obj->a, cl_obj->color, lv_obj->threshold, lv_obj->colorMax);


	/* ҥȥκ */
	cl_obj->histogram = (int *)CUDA_MALLOC_HOST(sizeof(int) * cl_obj->colorNum, "histogram");
	SetHistogram(cl_obj->colorNum, lv_obj->a.n, cl_obj->histogram, cl_obj->color);
	PrintHistogram(cl_obj->colorNum, cl_obj->histogram);


	/* 󥰤 */
	lv_obj->ordering = SetOrdering(cl_obj, lv_obj->a.np);

	/* BMP */
//	PrintBMP(&lv_obj->a, &lv_obj->inport_t, &lv_obj->export_t, lv_obj->ordering, MPI_GET_RANK());

	/* ࡼ */
	switch(lv_obj->smootherType){
	case JC:
		lv_obj->Smoother = CUDA_Jacobi;
		break;
	case GS:
		lv_obj->Smoother = CUDA_GaussSeidel;
		break;
	case JCGS:
		lv_obj->Smoother = CUDA_JCGS;
		break;
	case IDR:
		lv_obj->Smoother = CUDA_IDR_Jacobi;
		break;
	case LUD:
		int n;
		n  = lv_obj->a.n;
		l  = lv_obj->l = (double *)CUDA_MALLOC_HOST(sizeof(double) * n*n, "l");
		u  = lv_obj->u = (double *)CUDA_MALLOC_HOST(sizeof(double) * n*n, "u");
		LU_Decomposition(&lv_obj->a, l, u);
		lv_obj->Smoother = LU;
		break;
	case CG:
		lv_obj->Smoother = CUDA_CG;
		break;
	default:
		throw("please select the correct smoother.");
	}

	/*  */
	ordering_col = obj->level[lv].ordering;
	ordering_row = obj->level[lv].ordering;
	cl_obj = &obj->level[lv].cl_obj;
	SendLU_CRS(&lv_obj->a, lv_obj->da, cl_obj, ordering_col, ordering_row);

	if(lv!=0){
		ordering_col = obj->level[lv-1].ordering;
		ordering_row = obj->level[ lv ].ordering;
		cl_obj = &obj->level[ lv ].cl_obj;
		SendCRS(&lv_obj->r, lv_obj->dr, cl_obj, ordering_col, ordering_row);

		ordering_col = obj->level[ lv ].ordering;
		ordering_row = obj->level[lv-1].ordering;
		cl_obj = &obj->level[lv-1].cl_obj;
		SendCRS(&lv_obj->p, lv_obj->dp, cl_obj, ordering_col, ordering_row);
	}

	size = (int)lv_obj->da.size();
	if(size > 0){
		np = 0;
		for(i=0; i<size; i++){
			np += lv_obj->da[i].np;
		}
		lv_obj->n  = np - (lv_obj->da[size-1].np - lv_obj->da[size-1].n);
		lv_obj->np = lv_obj->n + lv_obj->a.np - lv_obj->a.n;

	}else{
		lv_obj->n  = 0;
		lv_obj->np = lv_obj->a.np;
	}

	DebugPrintf0("AMG SETUP End.\n");
	return;
}


void AMG_CYCLE_SETUP(object *obj, double *x, double *b)
{
	DebugPrintf0("AMG CYCLE SETUP Start.\n");
	int				i, levelNum, n, np;
	int				*ordering;
	double			*tempType;
	level_object	*lv_obj;

	levelNum = obj->levelNum;
	for(i=0; i<levelNum; i++){
		lv_obj = &obj->level[i];

		np = lv_obj->np;

		CUDA_MALLOC((void **)&lv_obj->dx, sizeof(double) * np, "dx malloc");
		CUDA_MALLOC((void **)&lv_obj->db, sizeof(double) * np, "db malloc");
		CUDA_MALLOC((void **)&lv_obj->de, sizeof(double) * np, "de malloc");

		CUDA_MEMSET(lv_obj->dx, 0, sizeof(double) * np);
		CUDA_MEMSET(lv_obj->db, 0, sizeof(double) * np);
		CUDA_MEMSET(lv_obj->de, 0, sizeof(double) * np);
	}

	lv_obj = &obj->level[0];
	tempType = (double *)CUDA_MALLOC_HOST(sizeof(double) * lv_obj->np, "amg cycle setup tempType");
	ordering = lv_obj->ordering;

	n = lv_obj->a.n;

	memset(tempType, 0, sizeof(double) * lv_obj->np);
	for(i=0; i<n; i++){
		tempType[ordering[i]] = x[i];
	}
	CUDA_SEND(lv_obj->dx, tempType, sizeof(double) * lv_obj->np);

	memset(tempType, 0, sizeof(double) * lv_obj->np);
	for(i=0; i<n; i++){
		tempType[ordering[i]] = b[i];
	}
	CUDA_SEND(lv_obj->db, tempType, sizeof(double) * lv_obj->np);

	CUDA_FREE_HOST(tempType);

	DebugPrintf0("AMG CYCLE SETUP End.\n");
	return;
}


void AMG_CYCLE(object *obj, double *dx, double *db, int *iter, int flag)
{
	DebugPrintf0("AMG CYCLE Start.\n");
	int				i, levelNum, loop, converge;
	double			*temp_dx, *temp_db;
	level_object	*lv_obj, *front, *rear;

	levelNum = obj->levelNum;

	lv_obj = &obj->level[0];
	temp_dx = lv_obj->dx;
	temp_db = lv_obj->db;
	lv_obj->dx = dx;
	lv_obj->db = db;

	if(flag){
		obj->first_residual = CUDA_Dot(lv_obj->db, lv_obj->db, lv_obj->n);
	}

	loop=0;
	do{
		loop++;
		/*
		lv_obj = &obj->level[0];
		lv_obj->Smoother(lv_obj);
		*/
//		/*

		for(i=0; i<levelNum-1; i++){
			DebugPrintf1("LEVEL %d\n", i);
			rear  = &obj->level[ i ];
			front = &obj->level[i+1];

			DebugPrintf0("Smoother.\n");
			rear->Smoother(rear);

			DebugPrintf0("Residual Error.\n");
			CUDA_ResidualError(rear->da, rear->dx, rear->db, rear->de);

			DebugPrintf0("Transition Matrix.\n");
			CUDA_TransitionMatrix(front->dr, rear->de, front->db);

			CUDA_MEMSET(front->dx, 0, sizeof(double) * front->np);
		}

		DebugPrintf1("LEVEL %d\n", levelNum-1);
		lv_obj = &obj->level[levelNum-1];

		DebugPrintf0("Smoother.\n");
		lv_obj->Smoother(lv_obj);


		for(i=levelNum-1; i>0; i--){
			DebugPrintf1("LEVEL %d\n", i);
			rear  = &obj->level[ i ];
			front = &obj->level[i-1];

			DebugPrintf0("ReTransition Matrix.\n");
			CUDA_ReTransitionMatrix(rear->dp, rear->dx, front->dx);

			DebugPrintf0("Smoother.\n");
			front->Smoother(front);
		}

//		*/

		if(flag){
			printf("loop:%3d ", loop);
			converge = obj->Terminal(obj);
		}
	}while(converge == false && loop < *iter && flag);
	*iter = loop;
	if(flag){
		printf("\n");
	}

	lv_obj = &obj->level[0];
	lv_obj->dx = temp_dx;
	lv_obj->db = temp_db;

	DebugPrintf0("AMG CYCLE End.\n");
	return;
}


void AMG_CYCLE_FINALIZE(object *obj, double *x, double *b)
{
	DebugPrintf0("AMG CYCLE FINALIZE Start.\n");
	int				i, levelNum, np;
	int				*ordering;
	double			*tempType;
	level_object	*lv_obj;

	levelNum = obj->levelNum;

	lv_obj = &obj->level[0];
	np = lv_obj->np;

	tempType = (double *)CUDA_MALLOC_HOST(sizeof(double) * np, "amg cycle finalize tempType");
	ordering = lv_obj->ordering;

	CUDA_LOAD(tempType, lv_obj->dx, sizeof(double) * np);

	np = lv_obj->a.np;

	for(i=0; i<np; i++){
		x[i] = tempType[ordering[i]];
	}

	CUDA_FREE_HOST(tempType);

	for(i=0; i<levelNum; i++){
		lv_obj = &obj->level[i];

		CUDA_FREE(lv_obj->dx);
		CUDA_FREE(lv_obj->db);
		CUDA_FREE(lv_obj->de);
	}

	DebugPrintf0("AMG CYCLE FINALIZE End.\n");
}


void LU_Decomposition(lu_crs *lu_a, double *a_l, double *a_u)
{
	DebugPrintf0("LU Decomposition Start.\n");

	int		i, j, k, index, np;
	double	val;
	double	*a, *l, *u;

	np = lu_a->np;
	a = (double *)CUDA_MALLOC_HOST(sizeof(double) * np*np, "lu a");
	l = (double *)CUDA_MALLOC_HOST(sizeof(double) * np*np, "lu l");
	u = (double *)CUDA_MALLOC_HOST(sizeof(double) * np*np, "lu u");

	/* ¹ */
	UnpackLU_CRS(lu_a, a);

	/* LUʬ */
	for(i=0; i<np; i++){
		for(j=0; j<np; j++){
			MAT(l, np, np, j, i) = 0.0;
			MAT(u, np, np, j, i) = 0.0;
		}
		MAT(l, np, np, i, i) = 1.0;
	}
	for(j=0; j<np; j++){
		MAT(u, np, np, j, 0) = MAT(a, np, np, j, 0);

		for(i=0; i<=j; i++){
			val = MAT(a, np, np, j, i);
			for(k=0; k<i; k++){
				val -= MAT(l, np, np, k, i) * MAT(u, np, np, j, k);
			}
			MAT(u, np, np, j, i) = val;
		}

		for(i=j+1; i<np; i++){
			val = MAT(a, np, np, j, i);
			for(k=0; k<j; k++){
				val -= MAT(l, np, np, k, i) * MAT(u, np, np, j, k);
			}
			MAT(l, np, np, j, i) = val / MAT(u, np, np, j, j);
		}
	}

	/* L */
	index = 0;
	for(i=0; i<np; i++){
		for(j=0; j<i; j++){
			a_l[index++] = MAT(l, np, np, j, i);
		}
		a_l[index++] = MAT(l, np, np, i, i);
	}

	/* U */
	index = 0;
	for(i=np-1; i>=0; i--){
		for(j=i+1; j<np; j++){
			a_u[index++] = MAT(u, np, np, j, i);
		}
		a_u[index++] = MAT(u, np, np, i, i);
	}

	CUDA_FREE_HOST(a), CUDA_FREE_HOST(l), CUDA_FREE_HOST(u);

	DebugPrintf0("LU Decomposition End.\n");
	return;
}


void UnpackLU_CRS(lu_crs *lu_a, double *a)
{
	int		i, j, start, end, np;
	int		*inl, *inu, *ial, *iau;
	double	*al, *au, *d;

	np	= lu_a->np;
	inl	= lu_a->inl;
	inu	= lu_a->inu;
	ial	= lu_a->ial;
	iau	= lu_a->iau;
	al	= lu_a->al;
	au	= lu_a->au;
	d	= lu_a->d;

	for(i=0; i<np; i++){
		for(j=0; j<np; j++){
			MAT(a, np, np, j, i) = 0.0;
		}
	}

	for(i=0; i<np; i++){

		start = (i) ? inl[i-1] : 0;
		end   = inl[i];
		for(j=start; j<end; j++){
			MAT(a, np, np, ial[j], i) = al[j];
		}

		start = (i) ? inu[i-1] : 0;
		end   = inu[i];
		for(j=start; j<end; j++){
			MAT(a, np, np, iau[j], i) = au[j];
		}

		MAT(a, np, np, i, i) = d[i];
	}

	return;
}


void LU(level_object *lv_obj){
	int i, np, dnp;
	int *ordering;
	double *l, *u, *b, *x, *db, *dx, *temp;

	np = lv_obj->a.np;
	dnp = lv_obj->np;
	ordering = lv_obj->ordering;
	l = lv_obj->l;
	u = lv_obj->u;
	db = lv_obj->db;
	dx = lv_obj->dx;

	x    = (double *)CUDA_MALLOC_HOST(sizeof(double) * np,  "lu x");
	b    = (double *)CUDA_MALLOC_HOST(sizeof(double) * np,  "lu b");
	temp = (double *)CUDA_MALLOC_HOST(sizeof(double) * dnp, "lu temp");

	CUDA_LOAD(temp, db, sizeof(double) * dnp);
	for(i=0; i<np; i++){
		b[i] = temp[ordering[i]];
	}

	ForwardElimination(l, b, x, np);
	BackwardSubstitution(u, b, x, np);

	for(i=0; i<np; i++){
		temp[ordering[i]] = x[i];
	}
	CUDA_SEND(dx, temp, sizeof(double) * dnp);

	CUDA_FREE_HOST(x);
	CUDA_FREE_HOST(b);
	CUDA_FREE_HOST(temp);

	return;
}


void ForwardElimination(double *l, double *b, double *x, int np)
{
	int i, j;
	int index;
	double val;

	index = 0;
	for(i=0; i<np; i++){
		val = b[i];
		for(j=0; j<i; j++){
			val -= l[index++] * x[j];
		}
		x[i] = val / l[index++];
	}
	return;
}


void BackwardSubstitution(double *u, double *b, double *x, int np)
{
	int i, j;
	int index;
	double val;

	index = 0;
	for(i=np-1; i>=0; i--){
		val = x[i];
		for(j=i+1; j<np; j++){
			val -= u[index++] * x[j];
		}	
		x[i] = val / u[index++];
	}
	return;
}


int SetColor(lu_crs *a, int *color, double threshold, int num)
{
	DebugPrintf0("SET COLOR Start.\n");
	int    i, pallet, n;
	bool   flag;
	double *max;

	n  = a->n;

	for(i=0; i<n; i++){
		color[i] = 0;
	}

	/* ³õ */
	max = (double *)CUDA_MALLOC_HOST(sizeof(double) * n, "max edge");
	MaxRowVal(a, max);

	/* դ */
	pallet = 0;
	do{
		pallet++;
		flag = PaintColor(a, pallet, color, max, threshold);
	}while(flag);

	/* 0οֹ˽ */
	for(i=0; i<n; i++){
		color[i]--;
	}
	pallet--;

	if(num != 0 && pallet>num){
		for(i=0; i<n; i++){
			color[i] %= num;
		}
		pallet = num;
	}

	CUDA_FREE_HOST(max);

	DebugPrintf0("SET COLOR Start.\n");
	return pallet;
}


void MaxRowVal(lu_crs *a, double *max)
{
	int		i, j, start, end, n;
	int		*inl, *inu;
	double	val;
	double	*al, *au;

	n	= a->n;
	inl	= a->inl;
	inu	= a->inu;
	al	= a->al;
	au	= a->au;

	for(i=0; i<n; i++){
		max[i] = 0.0;

		start = (i) ? inl[i-1] : 0;
		end   = inl[i];

		for(j=start; j<end; j++){
			val = fabs(al[j]);
			max[i] = val > max[i] ? val : max[i];
		}

		start = (i) ? inu[i-1] : 0;
		end   = inu[i];
		for(j=start; j<end; j++){
			val = fabs(au[j]);
			max[i] = val > max[i] ? val : max[i];
		}
	}
	return;
}


bool PaintColor(lu_crs *a, int pallet, int *color, double *max, double threshold)
{
	int		i, j, start, end, look, index, n;
	int		*inl, *inu, *ial, *iau;
	bool	flag;
	double	val, threshold2;
	double	*al, *au, *d;

	n	= a->n;
	inl	= a->inl;
	inu	= a->inu;
	ial	= a->ial;
	iau	= a->iau;
	al	= a->al;
	au	= a->au;
	d	= a->d;

	flag = false;
	for(i=0; i<n; i++){
		if(color[i] == 0){
			flag = true;
			threshold2 = threshold * max[i];

			start = (i) ? inl[i-1] : 0;
			end   = inl[i];
			for(j=start; j<end; j++){
				val = fabs(al[j]);
				index = ial[j];
				if(index < n){
					look = color[index];
				}else{
					look = 0;
				}

				if(val >= threshold2 && look == pallet){
					goto _next_node;/* ɤ餺˼ */
				}
			}

			start = (i) ? inu[i-1] : 0;
			end   = inu[i];
			for(j=start; j<end; j++){
				val = fabs(au[j]);
				index = iau[j];
				if(index < n){
					look = color[index];
				}else{
					look = 0;
				}

				if(val >= threshold2 && look == pallet){
					goto _next_node;/* ɤ餺˼ */
				}
			}

			color[i] = pallet;
		}

		_next_node:;
	}

	return flag;
}


void SetHistogram(int m, int n, int *histogram, int *data)
{
	DebugPrintf0("Set Histogram Start.\n");
	int i, temp;

	if(m > 0){
		for(i=0; i<m; i++){
			histogram[i] = 0;
		}

		for(i=0; i<n; i++){
			temp = data[i];
			histogram[temp]++;
		}
	}
	DebugPrintf0("Set Histogram End.\n");
	return;
}


void PrintHistogram(int n, int *histogram)
{
	DebugPrintf0("Print Histogram Start.\n");
	int i, j, total, raito;

	total = 0;
	for(i=0; i<n; i++){
		total += histogram[i];
	}

	//	printf("Total:%-8d Series:%-4d\n", total, n);

	for(i=0; i<n; i++){
	  //		printf("No.%2d:", i);
		raito = (int)((double)(64 * histogram[i]) / (double)(total));

		for(j=0; j<raito; j++){
		  //			printf("*");
		}

		//		printf("(%5d)\n", histogram[i]);
	}
	DebugPrintf0("Print Histogram End.\n");
	return;
}

/*
#define HEIGHT 512
#define WIDTH  512
void PrintBMP(lu_crs *a, int *ordering, int rank)
{
	int		i, j, x, y, start, end, index, n;
	int		*inl, *inu, *ial, *iau;
	double	val;
	double	*al, *au, *d;
	char fileName[128];

	img *bmp_data;
	bmp_data = (img *)malloc(sizeof(img));
	bmp_data->height = HEIGHT;
	bmp_data->width  = WIDTH;

	for(i=0; i<bmp_data->height; i++){
		for(j=0; j<bmp_data->width; j++){
			bmp_data->data[i][j].r = bmp_data->data[i][j].g
				                   = bmp_data->data[i][j].b 
								   = 0;
		}
	}

	n	= a->n;
	inl	= a->inl;
	inu	= a->inu;
	ial	= a->ial;
	iau	= a->iau;
	al	= a->al;
	au	= a->au;
	d	= a->d;

	for(i=0; i<n; i++){
		start = (i) ? inl[i-1] : 0;
		end   = inl[i];
		for(j=start; j<end; j++){
			val = fabs(al[j] / d[i]) * (end - start);
			val = val > 1 ? 1 : val;
			index = ial[j];

			double temp_x, temp_y;
			temp_x = (double)ordering[index] / n;
			temp_y = (double)ordering[  i  ] / n;

			if(temp_x < 1 && temp_y < 1){
				x = temp_x * (WIDTH-1);
				y = temp_y * (HEIGHT-1);
				bmp_data->data[y][x].r = bmp_data->data[y][x].g
				                   = bmp_data->data[y][x].b 
								   = (unsigned char)(255.0 * val);
			}
		}

		start = (i) ? inu[i-1] : 0;
		end   = inu[i];
		for(j=start; j<end; j++){
			val = fabs(au[j] / d[i]) * (end - start);
			val = val > 1 ? 1 : val;
			index = iau[j];

			double temp_x, temp_y;
			temp_x = (double)ordering[index] / n;
			temp_y = (double)ordering[  i  ] / n;

			if(temp_x < 1 && temp_y < 1){
				x = temp_x * (WIDTH-1);
				y = temp_y * (HEIGHT-1);
				bmp_data->data[y][x].r = bmp_data->data[y][x].g
				                   = bmp_data->data[y][x].b 
								   = (unsigned char)(255.0 * val);
			}

		}

	}
	sprintf(fileName, "after_%d_%d.bmp", rank, n);
	WriteBmp(fileName, bmp_data);


	for(i=0; i<bmp_data->height; i++){
		for(j=0; j<bmp_data->width; j++){
			bmp_data->data[i][j].r = bmp_data->data[i][j].g
				                   = bmp_data->data[i][j].b 
								   = 0;
		}
	}

	free(bmp_data);

	return;
}
*/

int *SetOrdering(color_object *cl_obj, int np)
{
	DebugPrintf0("Set Ordering Start.\n");
	int i, unit, partition, length, pallet;
	int *histogram, *row, *color, *ordering;

	ordering = (int *)CUDA_MALLOC_HOST(sizeof(int) * np, "ordering");

	unit = PADDING_BYTE / sizeof(int);

	partition = cl_obj->colorNum;
	histogram = cl_obj->histogram;
	length    = cl_obj->length;
	color     = cl_obj->color;


	if(partition > 0){
		row = (int *)CUDA_MALLOC_HOST(sizeof(int) * partition, "row");
		for(i=0; i<partition; i++){
			row[i] = (i) ? row[i-1] + (histogram[i-1] + (unit - histogram[i-1] % unit)) : 0;
		}

		for(i=0; i<length; i++){
			pallet = color[i];
			ordering[i] = row[pallet];
			row[pallet]++;
		}

		pallet = partition-1;
		for(i=length; i<np; i++){
			ordering[i] = row[pallet];
			row[pallet]++;
		}

		CUDA_FREE_HOST(row);

	}else{
		for(i=0; i<np; i++){
			ordering[i] = i;
		}
	}

	DebugPrintf0("Set Ordering End.\n");
	return ordering;
}


void SendLU_CRS(lu_crs *a, lu_icrss &da, color_object *cl_obj, int *ordering_col, int *ordering_row)
{
	DebugPrintf0("Send LU_CRS Start.\n");


	lu_icrs *dap;
	int i, j, colorNum, unit, np, nplu, num, temp, pallet, start, end, row, index;
	int *histogram, *inl, *inu, *ial, *iau;
	int *intBuf, **intTemp;
	double *doubleBuf, **doubleTemp, *al, *au, *d;

	unit = PADDING_BYTE / sizeof(int);

	colorNum  = cl_obj->colorNum;
	histogram = cl_obj->histogram;

	inl = a->inl;
	inu = a->inu;
	ial = a->ial;
	iau = a->iau;
	al  = a->al;
	au  = a->au;
	d   = a->d;

	da.resize(colorNum);
	for(i=0; i<colorNum; i++){
		da[i].original = true;
	}

	for(i=0; i<colorNum; i++){
		dap = &da[i];

		dap->n    = histogram[i];
		dap->np   = histogram[i] + (unit - histogram[i] % unit);
		dap->nplu = 0;
		dap->row  = i ? da[i-1].row + da[i-1].np : 0;
	}

	for(i=0; i<a->n; i++){
		dap = &da[ cl_obj->color[i] ];

		num = (i ? inl[i]-inl[i-1] + inu[i]-inu[i-1] : inl[i] + inu[i]) * dap->np;

		dap->nplu = num > dap->nplu ? num : dap->nplu;
	}


	temp = 0;
	for(i=0; i<colorNum; i++){
		dap = &da[i];

		np   = dap->np;
		nplu = dap->nplu;
		CUDA_MALLOC((void**)&dap->cnt, sizeof(int)    * np  , "lu cnt malloc");
		CUDA_MALLOC((void**)&dap->col, sizeof(int)    * nplu, "lu col malloc");
		CUDA_MALLOC((void**)&dap->val, sizeof(double) * nplu, "lu val malloc");
		CUDA_MALLOC((void**)&dap->d,   sizeof(double) * np  , "lu  d  malloc");

		temp += nplu;

	}
	intBuf     = (int*)CUDA_MALLOC_HOST(sizeof(int) * temp, "send lu crs intBuf");
	intTemp    = (int**)CUDA_MALLOC_HOST(sizeof(int*) * colorNum, "send lu crs intTemp");
	doubleBuf  = (double*)CUDA_MALLOC_HOST(sizeof(double) * temp, "send lu crs doubleBuf");
	doubleTemp = (double**)CUDA_MALLOC_HOST(sizeof(double*) * colorNum, "send lu crs doubleTemp");


	memset(intBuf, 0, sizeof(int) * temp);
	memset(doubleBuf, 0, sizeof(double) * temp);
	for(i=0; i<a->n; i++){
		index = ordering_row[i];
		intBuf[index] = i ? inl[i]-inl[i-1] + inu[i]-inu[i-1] : inl[i] + inu[i];
		doubleBuf[index] = d[i];
	}

	temp = 0;
	for(i=0; i<colorNum; i++){
		intTemp[i]    = intBuf    + temp;
		doubleTemp[i] = doubleBuf + temp;

		dap = &da[i];
		CUDA_SEND(dap->cnt, intBuf + dap->row, sizeof(int) * dap->np);
		CUDA_SEND(dap->d, doubleBuf + dap->row, sizeof(double) * dap->np);
	
		nplu = dap->nplu;
		temp += nplu;
	}


	memset(intBuf, 0, sizeof(int) * temp);
	memset(doubleBuf, 0, sizeof(double) * temp);
	for(i=0; i<a->n; i++){
		pallet = cl_obj->color[i];
		row    = ordering_row[i];
		dap    = &da[pallet];
	
		index  = row - dap->row;

		start  = i ? inl[i-1] : 0;
		end    = inl[i];
		for(j=start; j<end; j++){
			intTemp[pallet][index]    = ordering_col[ial[j]];
			doubleTemp[pallet][index] = al[j];
			index += dap->np;
		}

		start  = i ? inu[i-1] : 0;
		end    = inu[i];
		for(j=start; j<end; j++){
			intTemp[pallet][index]    = ordering_col[iau[j]];
			doubleTemp[pallet][index] = au[j];
			index += dap->np;
		}
	}

	for(i=0; i<colorNum; i++){
		dap = &da[i];
		CUDA_SEND(dap->col, intTemp[i],    sizeof(int)    * dap->nplu);
		CUDA_SEND(dap->val, doubleTemp[i], sizeof(double) * dap->nplu);
	}

	CUDA_FREE_HOST(intBuf);
	CUDA_FREE_HOST(intTemp);
	CUDA_FREE_HOST(doubleBuf);
	CUDA_FREE_HOST(doubleTemp);

	DebugPrintf0("Send LU_CRS End.\n");
	return;
}


void SendCRS(crs *a, icrss &da, color_object *cl_obj, int *ordering_col, int *ordering_row)
{
	DebugPrintf0("Send CRS Start.\n");

	icrs *dap;
	int i, j, colorNum, unit, np, nplu, num, temp, pallet, start, end, row, index, dummy;
	int *histogram, *dsp, *col;
	int *intBuf, **intTemp;
	double *doubleBuf, **doubleTemp, *val;

	if(a->nplu <= 0){
		da.resize(0);
		return;
	}

	unit = PADDING_BYTE / sizeof(int);

	colorNum  = cl_obj->colorNum;
	histogram = cl_obj->histogram;

	dsp = a->dsp;
	col = a->col;
	val = a->val;

	if(colorNum == 0){
		colorNum = 1;
		histogram = &dummy;
		histogram[0] = 0;
	}

	da.resize(colorNum);
	for(i=0; i<colorNum; i++){
		da[i].original = true;
	}

	histogram[colorNum-1] += a->np - a->n;
	for(i=0; i<colorNum; i++){
		dap = &da[i];

		dap->n    = histogram[i];
		dap->np   = histogram[i] + (unit - histogram[i] % unit);
		dap->nplu = 0;
		dap->row  = i ? da[i-1].row + da[i-1].np : 0;
	}
	histogram[colorNum-1] -= a->np - a->n;

	for(i=0; i<a->n; i++){
		dap = &da[ cl_obj->color[i] ];

		num = (i ? dsp[i] - dsp[i-1] : dsp[i]) * dap->np;

		dap->nplu = num > dap->nplu ? num : dap->nplu;
	}
	for(i=a->n; i<a->np; i++){
		dap = &da[ colorNum-1 ];

		num = (i ? dsp[i] - dsp[i-1] : dsp[i]) * dap->np;

		dap->nplu = num > dap->nplu ? num : dap->nplu;
	}

	temp = 0;
	for(i=0; i<colorNum; i++){
		dap = &da[i];

		np   = dap->np;
		nplu = dap->nplu;

		CUDA_MALLOC((void**)&dap->cnt, sizeof(int)    * np  , "cnt malloc");
		CUDA_MALLOC((void**)&dap->col, sizeof(int)    * nplu, "col malloc");
		CUDA_MALLOC((void**)&dap->val, sizeof(double) * nplu, "val malloc");

		CUDA_MEMSET(dap->cnt, 0, sizeof(int)    * np);
		CUDA_MEMSET(dap->col, 0, sizeof(int)    * nplu);
		CUDA_MEMSET(dap->val, 0, sizeof(double) * nplu);

		temp += nplu;

	}
	intBuf     = (int*)CUDA_MALLOC_HOST(sizeof(int) * temp, "send crs intBuf");
	intTemp    = (int**)CUDA_MALLOC_HOST(sizeof(int*) * colorNum, "send crs intTemp");
	doubleBuf  = (double*)CUDA_MALLOC_HOST(sizeof(double) * temp, "send crs doubleBuf");
	doubleTemp = (double**)CUDA_MALLOC_HOST(sizeof(double*) * colorNum, "send crs doubleTemp");


	memset(intBuf, 0, sizeof(int) * temp);
	for(i=0; i<a->np; i++){
		intBuf[ordering_row[i]] = i ? dsp[i] - dsp[i-1] : dsp[i];
	}

	temp = 0;
	for(i=0; i<colorNum; i++){
		intTemp[i]    = intBuf    + temp;
		doubleTemp[i] = doubleBuf + temp;

		dap = &da[i];
		CUDA_SEND(dap->cnt, intBuf + dap->row, sizeof(int) * dap->np);
	
		nplu = dap->nplu;
		temp += nplu;
	}


	memset(intBuf, 0, sizeof(int) * temp);
	memset(doubleBuf, 0, sizeof(double) * temp);
	for(i=0; i<a->n; i++){
		pallet = cl_obj->color[i];
		start  = (i) ? dsp[i-1] : 0;
		end    = dsp[i];
		row    = ordering_row[i];

		dap   = &da[pallet];
		index = row - dap->row;
		for(j=start; j<end; j++){
			intTemp[pallet][index]    = ordering_col[col[j]];
			doubleTemp[pallet][index] = val[j];
			index += dap->np;
		}
	}
	for(i=a->n; i<a->np; i++){
		pallet = colorNum-1;
		start  = (i) ? dsp[i-1] : 0;
		end    = dsp[i];
		row    = ordering_row[i];

		dap   = &da[pallet];
		index = row - dap->row;
		for(j=start; j<end; j++){
			intTemp[pallet][index]    = ordering_col[col[j]];
			doubleTemp[pallet][index] = val[j];
			index += dap->np;
		}
	}

	for(i=0; i<colorNum; i++){
		dap = &da[i];
		CUDA_SEND(dap->col, intTemp[i],    sizeof(int)    * dap->nplu);
		CUDA_SEND(dap->val, doubleTemp[i], sizeof(double) * dap->nplu);
	}

	CUDA_FREE_HOST(intBuf);
	CUDA_FREE_HOST(intTemp);
	CUDA_FREE_HOST(doubleBuf);
	CUDA_FREE_HOST(doubleTemp);

	DebugPrintf0("Send CRS End.\n");
	return;
}
