/*
   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
   3. Neither the name of the project nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "gpu_solver.h"
#include "AMG.h"
#include "BiCGSTAB.h"
#include "mySMOOTHER.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

enum SOLVER_TYPE{AMG, BiCGSTAB, SOLVER_LENGTH};

object *gpu_solver_obj;
int solver_select;

template <class T>
T *SendVector(T *vector, int vector_length, int device_length, int *ordering);



void notify_level_(int *Level_NUM)
{
	CUDA_SETUP(0);
	gpu_solver_obj = new object;
	gpu_solver_obj->levelNum = *Level_NUM;
	gpu_solver_obj->level.reserve(*Level_NUM);

	gpu_solver_obj->residual  = 1.0e-4;
	gpu_solver_obj->threshold = NULL;

	return;
}


//void parameter_set_on_gpu_(int *select, double *omega)
void parameter_set_on_gpu_(int *select, double *omega, int* iter_down, int* iter_bottom, int* gs_at_lev1)
{

	int i, levelNum;
	level_object *lv_obj;

	solver_select = *select;

	levelNum = gpu_solver_obj->levelNum;

	for(i=0; i<levelNum; i++){
		/* ʹ߲ */
		if( i < (int)gpu_solver_obj->level.size() ){
			level_objects::iterator it = gpu_solver_obj->level.begin();
			gpu_solver_obj->level.insert(it+i, level_object());
		}else{
			gpu_solver_obj->level.push_back(level_object());
		}

		lv_obj = &gpu_solver_obj->level[i];

		// set smoother type.
		if(i==0){
		  // lv_obj->smootherType = GS;
			if(*gs_at_lev1) lv_obj->smootherType = GS;//added
			else lv_obj->smootherType = JC;//added
		}else if(i==levelNum-1){
			lv_obj->smootherType = LUD;
		}else{
			lv_obj->smootherType = JC;
		}

		// set number of trial
		if(lv_obj->smootherType == LUD){
			lv_obj->numberOfTrials = 1;
		}else{
			if(i!=levelNum-1){
				lv_obj->numberOfTrials = 2;
				lv_obj->numberOfTrials = *iter_down;//added
			}else{
				lv_obj->numberOfTrials = 30;
				lv_obj->numberOfTrials = *iter_bottom;//added
			}
		}

		// set limit number of color.
		if(lv_obj->smootherType != GS && lv_obj->smootherType != JCGS ){
			lv_obj->colorMax = 1;
		}else{
			lv_obj->colorMax = 0;
		}

		// set threshold for multi color.
		if(i==0){
			lv_obj->threshold = 0.0;
		}else{
			lv_obj->threshold = 0.9;
		}

		// set omega.
		lv_obj->omega = *omega;
	}

	return;
}


void store_mat_on_gpu_(int *Level, 
					  int *N, int *NP, int *NPL, int *NPU, 
					  int *INL, int *INU, 
					  int *IAL, double *AL, 
					  int *IAU, double *AU, 
					  double *D,

					  int *R_N, int *P_N, int *R_NP, int *P_NP, int *R_NPLU, int *P_NPLU, 
					  int *R_IN, int *R_CN, double *R_V,
					  int *P_IN, int *P_CN, double *P_V)
{
	int i, npl, npu, nplu;
	level_object *lv_obj;
	lu_crs *a;
	crs *r, *p;

	lv_obj = &gpu_solver_obj->level[*Level-1];
	a = &lv_obj->a;
	r = &lv_obj->r;
	p = &lv_obj->p;

	a->n	= *N;
	a->np	= *NP;
	a->npl	= *NPL;
	a->npu	= *NPU;
	a->inl	= INL+1;
	a->inu	= INU+1;
	a->ial	= IAL;
	a->iau	= IAU;
	a->al	= AL;
	a->au	= AU;
	a->d	= D;

	if(*Level-1 != 0){
		r->n    = *R_N;
		r->np   = *R_NP;
		r->nplu = *R_NPLU;
		r->dsp  = R_IN+1;
		r->col  = R_CN;
		r->val  = R_V;

		p->n    = *P_N;
		p->np   = *P_NP;
		p->nplu = *P_NPLU;
		p->dsp  = P_IN+1;
		p->col  = P_CN;
		p->val  = P_V;
	}

	// modified col--
	npl = a->npl;
	npu = a->npu;
	for(i=0; i<npl; i++){
		a->ial[i]--;
	}
	for(i=0; i<npu; i++){
		a->iau[i]--;
	}
	if(*Level-1 != 0){
		nplu = r->nplu;
		for(i=0; i<nplu; i++){
			r->col[i]--;
		}

		nplu = p->nplu;
		for(i=0; i<nplu; i++){
			p->col[i]--;
		}
	}

	switch(solver_select){
	case AMG:
		AMG_SETUP(gpu_solver_obj, *Level-1);
		break;
	case BiCGSTAB:
		BiCGSTAB_SETUP(gpu_solver_obj, *Level-1);
		break;
	default:
		throw "please select AMG or BiCGSTAB !";
	}


	// modified col++
	npl = a->npl;
	npu = a->npu;
	for(i=0; i<npl; i++){
		a->ial[i]++;
	}
	for(i=0; i<npu; i++){
		a->iau[i]++;
	}
	if(*Level-1 != 0){
		nplu = r->nplu;
		for(i=0; i<nplu; i++){
			r->col[i]++;
		}

		nplu = p->nplu;
		for(i=0; i<nplu; i++){
			p->col[i]++;
		}
	}

	return;
}


void amgs_terminal_vec_set_on_gpu_(double *res, int *row)
{
	level_object *lv_obj;
	lv_obj = &gpu_solver_obj->level[0];

	gpu_solver_obj->restol = res;
	gpu_solver_obj->rowsize = *row;
	gpu_solver_obj->Terminal = CUDA_TerminalVec;

	if(gpu_solver_obj->threshold != NULL){
		CUDA_FREE(gpu_solver_obj->threshold);
	}
	gpu_solver_obj->threshold = SendVector(res, *row, lv_obj->np, lv_obj->ordering);

	return;
}


void amgs_terminal_val_set_on_gpu_(double *res)
{
	gpu_solver_obj->residual  = *res;
	gpu_solver_obj->threshold = &gpu_solver_obj->residual;
	gpu_solver_obj->Terminal  = CUDA_TerminalVal;
	return;
}

void solve_on_gpu_(double *x, double *b, int *iter)
{
	double *dx, *db;

	switch(solver_select){
	case AMG:
		AMG_CYCLE_SETUP(gpu_solver_obj, x, b);

		dx = gpu_solver_obj->level[0].dx;
		db = gpu_solver_obj->level[0].db;
		AMG_CYCLE(gpu_solver_obj, dx, db, iter, 1);

		AMG_CYCLE_FINALIZE(gpu_solver_obj, x, b);
		break;
	case BiCGSTAB:
		BiCGSTAB_CYCLE(gpu_solver_obj, x, b, iter);
		break;
	default:
		throw("please select solver!");
	}

	return;
}


void clear_mat_on_gpu_(void)
{

	delete gpu_solver_obj;

	//fflush is moved in this function.
	CUDA_LAST_ERROR_CHECK();

	return;
}

template <class T>
T *SendVector(T *vector, int vector_length, int device_length, int *ordering)
{
	int i;
	T *device_pointer, *temp;

	CUDA_MALLOC((void **)&device_pointer, sizeof(T) * device_length, "send vector pointer");
	temp = (T *)CUDA_MALLOC_HOST(sizeof(T) * device_length, "send vector temp");

	for(i=0; i<device_length; i++){
		temp[i] = HUGE_VAL;
	}
	for(i=0; i<vector_length; i++){
		temp[ordering[i]] = vector[i];
	}
	CUDA_SEND(device_pointer, temp, sizeof(T) * device_length);

	CUDA_FREE_HOST(temp);

	return device_pointer;
}
