/*
   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
   3. Neither the name of the project nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "myCUDA.h"
#include <stdio.h>
#include <stdlib.h>
#include <cutil.h>
#include <list>

class malloc_item
{
public:
	void *pointer;
	size_t size;
	char str[1024];

	malloc_item(void* pointer, const size_t size, const char str[])
	{
		int i;
		this->pointer = pointer;
		this->size    = size;

		for(i=0; str[i]; i++){
			this->str[i] = str[i];
		}
	}
};

int cuda_malloc_count = 0;
std::list<malloc_item> cuda_malloc_list;

int malloc_count = 0;
std::list<malloc_item> malloc_list;

void CUDA_INIT(int argc, char** argv)
{
	CUT_DEVICE_INIT(argc, argv);
	return;
}


void CUDA_EXIT(int argc, char** argv)
{
	CUT_EXIT(argc, argv);
	return;
}


void CUDA_SETUP(int rank)
{
	int device_num, n;
	cudaGetDeviceCount(&device_num);

	n = rank % device_num;

	cudaSetDevice(n);

	cudaGetDevice(&device_num);

//	printf("### cuda use %d device. ###\n", device_num);

	return;
}


void CUDA_SYNCHRONIZE(void)
{
	cudaThreadSynchronize();
	return;
}

void CUDA_MALLOC(void **devptr, size_t count, char str[])
{
	cudaError_t error_text;

	error_text = cudaMalloc(devptr, count);

	if (error_text != cudaSuccess) {
		printf("cuda malloc error : %s %u\n", cudaGetErrorString(error_text), count);
		throw;
	}

	cuda_malloc_count++;
	cuda_malloc_list.push_back( malloc_item(*devptr, count, str) );

	cudaMemset(*devptr, 0, count);

	return;
}


void *CUDA_MALLOC_HOST(size_t count, char str[])
{
	void *devptr;

	if(count > 0){
		devptr = malloc(count);
		if (devptr == NULL) {
			printf("not enough memory. size:%u\n", count);
			throw;
		}
	}else{
		devptr = NULL;
	}

	malloc_count++;
	malloc_list.push_back( malloc_item(devptr, count, str) );

	return devptr;
}


void CUDA_FREE(void *devptr)
{
	cudaError_t error_text;

	error_text = cudaFree(devptr);

	if (error_text != cudaSuccess) {
		printf("cuda free error : %s\n", cudaGetErrorString(error_text));
		throw;
	}
	cuda_malloc_count--;

	std::list<malloc_item>::iterator it;
	for(it=cuda_malloc_list.begin(); it!=cuda_malloc_list.end();){
		if((*it).pointer == devptr){
			it = cuda_malloc_list.erase(it);
			continue;
		}
		it++;
	}

	return;
}


void CUDA_FREE_HOST(void *devptr)
{
	if(devptr != NULL){
		free(devptr);
	}

	malloc_count--;

	std::list<malloc_item>::iterator it;
	for(it=malloc_list.begin(); it!=malloc_list.end();){
		if((*it).pointer == devptr){
			it = malloc_list.erase(it);
			continue;
		}
		it++;
	}
	return;
}


void CUDA_MEMSET(void *devptr, int c, size_t count)
{
	cudaMemset(devptr, c, count);
}


void CUDA_COPY(void *dst, const void *src, size_t count)
{
	cudaMemcpy(dst, src, count, cudaMemcpyDeviceToDevice);
	return;
}


void CUDA_SEND(void *dst, const void *src, size_t count)
{
	cudaMemcpy(dst, src, count, cudaMemcpyHostToDevice);
	return;
}


void CUDA_LOAD(void *dst, const void *src, size_t count)
{
	cudaMemcpy(dst, src, count, cudaMemcpyDeviceToHost);
	return;
}

void CUDA_LAST_ERROR_CHECK(void)
{
	cudaError_t err = cudaGetLastError();

	if (err != cudaSuccess) {
		printf("### cuda error : %s ###\n", cudaGetErrorString(err));
		fflush(stdout);
	}


	std::list<malloc_item>::iterator it;

	if(cuda_malloc_count){
		printf("cuda malloc count:%d\n", cuda_malloc_count);
		for(it=cuda_malloc_list.begin(); it!=cuda_malloc_list.end(); it++){
			printf("%p %8u %s\n", (*it).pointer, (*it).size, (*it).str);
		}
		fflush(stdout);
	}

	if(malloc_count){
		printf("malloc count:%d\n", malloc_count);
		for(it=malloc_list.begin(); it!=malloc_list.end(); it++){
			printf("%p %8u %s\n", (*it).pointer, (*it).size, (*it).str);
		}
		fflush(stdout);
	}
}


__global__ void CUDA_ADD_Kernel(double *a, double *b, double *c, double alpha, int n)
{
	unsigned int id = blockIdx.x * blockDim.x + threadIdx.x;
	if(id < n){
		c[id] = a[id] + alpha * b[id];
	}
}


void CUDA_ADD(double *a, double *b, double *c, double alpha, int n)
{
	CUDA_ADD_Kernel<<<n / 512 + 1, 512>>>(a, b, c, alpha, n);
	return;
}


__global__ void CUDA_SCALAR_MUL_Kernel(double a, double *b, double *c, int n)
{
	unsigned int id = blockIdx.x * blockDim.x + threadIdx.x;
	if(id < n){
		c[id] = a*b[id];
	}
}


void CUDA_SCALAR_MUL(double a, double *b, double *c, int n)
{
	CUDA_SCALAR_MUL_Kernel<<<n / 512 + 1, 512>>>(a, b, c, n);
	return;
}


__global__ void CUDA_EXTRACT_Kernel(double *dst, const double *src, const int *ordering, const int n)
{
	unsigned int id = blockIdx.x * blockDim.x + threadIdx.x;
	if(id < n){
		dst[id] = src[ ordering[id] ];
	}
}


void CUDA_EXTRACT(double *dst, const double *src, const int *ordering, const int n)
{
	CUDA_EXTRACT_Kernel<<<n / 512 + 1, 512>>>(dst, src, ordering, n);
	return;
}

__global__ void CUDA_SPRAY_Kernel(double *dst, const double *src, const int *ordering, const int n)
{
	unsigned int id = blockIdx.x * blockDim.x + threadIdx.x;
	if(id < n){
		dst[ ordering[id] ] += src[id];
	}
}


void CUDA_SPRAY(double *dst, const double *src, const int *ordering, const int n)
{
	CUDA_SPRAY_Kernel<<<n / 512 + 1, 512>>>(dst, src, ordering, n);
	return;
}
