/*
   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
   3. Neither the name of the project nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "mySMOOTHER.h"
#include "myCUDA.h"
#include "AMG_define_struct.h"

#define SP 256
#define SM 32

/*-----------------------------------------------------------------*/
template <unsigned int blockSize, class T>
__device__
void warpReduce(volatile T sdata[], int tid){
	if(blockSize >= 64) sdata[tid] += sdata[tid + 32];
	if(blockSize >= 32) sdata[tid] += sdata[tid + 16];
	if(blockSize >= 16) sdata[tid] += sdata[tid +  8];
	if(blockSize >=  8) sdata[tid] += sdata[tid +  4];
	if(blockSize >=  4) sdata[tid] += sdata[tid +  2];
	if(blockSize >=  2) sdata[tid] += sdata[tid +  1];
}

template <unsigned int blockSize>
__device__
void warpReduce(volatile bool sdata[], int tid){
	if(blockSize >= 64) sdata[tid] = sdata[tid] && sdata[tid + 32];
	if(blockSize >= 32) sdata[tid] = sdata[tid] && sdata[tid + 16];
	if(blockSize >= 16) sdata[tid] = sdata[tid] && sdata[tid +  8];
	if(blockSize >=  8) sdata[tid] = sdata[tid] && sdata[tid +  4];
	if(blockSize >=  4) sdata[tid] = sdata[tid] && sdata[tid +  2];
	if(blockSize >=  2) sdata[tid] = sdata[tid] && sdata[tid +  1];
}

/*
	ؿ̾: SharedReduction
	: T sdata[], int threadID
	֤: void
	ư: ɡ˳ݤ줿¤Ƭ˳Ǽ
*/
template <unsigned int blockSize, class T>
__device__
void SharedReduction(T sdata[], int tid)
{

	if(blockSize >= 512){ if(tid <256){ sdata[tid] += sdata[tid + 256];} __syncthreads();}
	if(blockSize >= 256){ if(tid <128){ sdata[tid] += sdata[tid + 128];} __syncthreads();}
	if(blockSize >= 128){ if(tid < 64){ sdata[tid] += sdata[tid +  64];} __syncthreads();}

	if(tid < 32) warpReduce<blockSize>(sdata, tid);
}

template <unsigned int blockSize>
__device__
void SharedReduction(bool sdata[], int tid)
{
	if(blockSize >= 512){ if(tid <256){ sdata[tid] = sdata[tid] && sdata[tid + 256];} __syncthreads();}
	if(blockSize >= 256){ if(tid <128){ sdata[tid] = sdata[tid] && sdata[tid + 128];} __syncthreads();}
	if(blockSize >= 128){ if(tid < 64){ sdata[tid] = sdata[tid] && sdata[tid +  64];} __syncthreads();}

	if(tid < 32) warpReduce<blockSize>(sdata, tid);
}


/*-----------------------------------------------------------------*/
__global__ void JCGSTask(
			lu_icrs a, double *old_x, double *new_x, double *b, double omega)
{
	__shared__ double s_val[SP];
	__shared__ double s_x[SP];
	__shared__ int    s_col[SP];

	int i, index, local_w;
	double temp, local_d, local_b, local_x;

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		index = id + a.row;
		local_w = a.cnt[id];
		local_d = a.d[id];
		local_x = old_x[index];
		local_b = b[index];

		temp = 0.0;

		index = id;
		for(i=0; i<local_w; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_x[tid] = old_x[s_col[tid]];

			temp -= s_val[tid] * s_x[tid];
		}

		temp += local_b;
		temp /= local_d;

		index = id + a.row;
		new_x[index] = local_x + omega * (temp - local_x);
	}
}


void CUDA_Jacobi(level_object *lv_obj)
{
	int i, j, size, numberOfTrials;
	double omega;
	double *dx, *temp_dx, *new_dx, *db, *temp;

	dx = lv_obj->dx;
	db = lv_obj->db;


	numberOfTrials = lv_obj->numberOfTrials;
	omega = lv_obj->omega;

	size = lv_obj->da.size();

	CUDA_MALLOC((void**)&temp_dx, sizeof(double)*lv_obj->np, "jc malloc");

	new_dx = temp_dx;
	for(i=0; i<numberOfTrials; i++){
		for(j=0; j<size; j++){
			JCGSTask<<< lv_obj->da[j].n / SP + 1, SP>>>
			(lv_obj->da[j], dx, new_dx, db, omega);
		}

		temp = dx, dx = new_dx, new_dx = temp;
	}

	if(dx != lv_obj->dx){
		CUDA_COPY(lv_obj->dx, dx, sizeof(double) * lv_obj->np);
	}

	CUDA_FREE(temp_dx);
	return;
}

void CUDA_GaussSeidel(level_object *lv_obj)
{
	int i, j, size, numberOfTrials;
	double omega;
	double *dx, *db;

	dx = lv_obj->dx;
	db = lv_obj->db;

	numberOfTrials = lv_obj->numberOfTrials;
	omega = lv_obj->omega;

	size = lv_obj->da.size();

	for(i=0; i<numberOfTrials; i++){
		for(j=0; j<size; j++){

			JCGSTask<<< lv_obj->da[j].n / SP + 1, SP>>>
			(lv_obj->da[j], dx, dx, db, omega);
		}
	}
	return;
}


void CUDA_JCGS(level_object *lv_obj)
{
	int i, j, size, numberOfTrials;
	double omega;
	double *dx, *temp_dx, *new_dx, *db, *temp;

	dx = lv_obj->dx;
	db = lv_obj->db;


	numberOfTrials = lv_obj->numberOfTrials;
	omega = lv_obj->omega;

	size = lv_obj->da.size();

	CUDA_MALLOC((void**)&temp_dx, sizeof(double)*lv_obj->np, "jcgs malloc");


	new_dx = temp_dx;
	for(i=0; i<numberOfTrials; i++){
		JCGSTask<<< lv_obj->da[0].n / SP + 1, SP>>>
		(lv_obj->da[0], dx, new_dx, db, omega);

		for(j=1; j<size; j++){
			JCGSTask<<< lv_obj->da[j].n / SP + 1, SP>>>
			(lv_obj->da[j], new_dx, new_dx, db, omega);
		}

		temp = dx, dx = new_dx, new_dx = temp;
	}

	if(dx != lv_obj->dx){
		CUDA_COPY(lv_obj->dx, dx, sizeof(double) * lv_obj->np);
	}

	CUDA_FREE(temp_dx);
	return;
}


__global__
void IDR_Kernel_0(
	double *s, double *d, double *r, double *dr, double *dx,
	double omega, double gamma, int n)
{
	const int id = blockIdx.x * blockDim.x + threadIdx.x;
	double temp;
	if(id < n){
		temp = (omega / d[id]) * (r[id] - gamma*dr[id]);
		s[id] = temp;
		dx[id] = temp - gamma * dx[id];
	}
}


__global__
void IDR_Kernel_1(lu_icrs a, double *s, double *r, double *dr, double omega)
{
	__shared__ double s_val[SP];
	__shared__ double s_s[SP];
	__shared__ int    s_col[SP];

	int i, index, local_w;
	double temp, local_d, local_s, local_r;

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		index = id;
		local_w = a.cnt[index];
		local_d = a.d[index];
		local_s = s[index];
		local_r = r[index];

		temp = 0.0;

		index = id;
		for(i=0; i<local_w; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_s[tid] = s[s_col[tid]];

			temp -= s_val[tid] * s_s[tid];
		}
		temp -= (1.0 - 1.0 / omega) * local_d * local_s;

		index = id;
		dr[index] = temp - local_r;
	}
}


void CUDA_IDR_Jacobi(level_object *lv_obj)
{
	int i, np, numberOfTrials;
	double omega, gamma;
	double *dx;
	double *buf, *dr, *dp, *ds, *ddr, *ddx;


	np = lv_obj->da[0].n;

	dx = lv_obj->dx;

	numberOfTrials = lv_obj->numberOfTrials;
	omega = lv_obj->omega;

	CUDA_MALLOC((void**)&buf, sizeof(double) * np * 5, "idr jc malloc");
	dr  = buf + 0 * np;
	dp  = buf + 1 * np;
	ds  = buf + 2 * np;
	ddr = buf + 3 * np;
	ddx = buf + 4 * np;

	CUDA_ResidualError(lv_obj->da, lv_obj->dx, lv_obj->db, dr);

	double *p;
	p = (double *)CUDA_MALLOC_HOST(sizeof(double) * np, "IDR Jacobi p");
	for(i=0; i<np; i++){
		p[i] = (double)rand() / (double)RAND_MAX;
	}
	CUDA_SEND(dp, p, sizeof(double) * np);
	CUDA_FREE_HOST(p);

	CUDA_COPY(ddr, dr, sizeof(double) * np);
	CUDA_COPY(ddx, dx, sizeof(double) * np);

	for(i=0; i<numberOfTrials; i++){
		gamma = i ? CUDA_Dot(dp, dr, np) / CUDA_Dot(dp, ddr, np) : 0.0;

		IDR_Kernel_0<<<np / SP + 1, SP>>>(ds, lv_obj->da[0].d, dr, ddr, ddx, omega, gamma, np);
		IDR_Kernel_1<<<np / SP + 1, SP>>>(lv_obj->da[0], ds, dr, ddr, omega);

		CUDA_ADD(dr, ddr, dr, 1.0, np);
		CUDA_ADD(dx, ddx, dx, 1.0, np);
	}

	CUDA_FREE(buf);

	return;
}
/*-----------------------------------------------------------------*/

void CUDA_CG(level_object *lv_obj)
{
	int i, np, numberOfTrials;
	double alpha, beta;
	double *buf, *dx, *dr, *dp, *dap;

	np = lv_obj->np;


	dx = lv_obj->dx;

	numberOfTrials = lv_obj->numberOfTrials;

	CUDA_MALLOC((void**)&buf, sizeof(double) * np * 3, "cg malloc");
	dr  = buf + 0 * np;
	dp  = buf + 1 * np;
	dap = buf + 2 * np;

	// r = b - Ax
	CUDA_ResidualError(lv_obj->da, lv_obj->dx, lv_obj->db, dr);

	// beta = 1 / rr
	beta = 1.0 / CUDA_Dot(dr, dr, np);

	// p = beta * r
	CUDA_SCALAR_MUL(beta, dr, dp, np);

	for(i=0; i<numberOfTrials; i++){
		// ap = a * p
		CUDA_MMULT(lv_obj->da, dp, dap);

		// alpha = 1 / pap
		alpha = 1.0 / CUDA_Dot(dp, dap, np);

		// x = x + alpha * p
		CUDA_ADD(dx, dp, dx, alpha, np);
		// r = r - alpha * ap
		CUDA_ADD(dr, dap, dr, -alpha, np);
		
		// beta 1 / rr
		beta = 1.0 / CUDA_Dot(dr, dr, np);

		// p = r + beta * p
		CUDA_ADD(dr, dp, dp, beta, np);
	}

	CUDA_FREE(buf);

	return;
}

/*-----------------------------------------------------------------*/

__global__
void ResidualErrorTask(lu_icrs a, double *x, double *b, double *r)
{
	__shared__ double s_val[SP];
	__shared__ double s_x[SP];
	__shared__ int    s_col[SP];

	int i, index, local_w;
	double temp, local_d, local_b, local_x;

	/* ᥤid,åidγ */
	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		/* d,b,xʬ򥰥Х꤫ƽ */
		index = id + a.row;
		local_w = a.cnt[id];
		local_d = a.d[id];
		local_x = x[index];
		local_b = b[index];

		temp = 0.0;

		index = id;
		for(i=0; i<local_w; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_x[tid] = x[s_col[tid]];

			temp -= s_val[tid] * s_x[tid];
		}
		temp -= local_d * local_x;

		index = id + a.row;
		r[index] = temp + local_b;
	}
}


void CUDA_ResidualError(lu_icrss &da, double *dx, double *db, double *de)
{
	int i, size;

	size = da.size();

	for(i=0; i<size; i++){
		ResidualErrorTask<<< da[i].n / SP + 1, SP>>>(da[i], dx, db, de);
	}
	return;
}


/*-----------------------------------------------------------------*/
__global__
void TransitionMatrixTask(icrs a, double *b, double *c)
{
	__shared__ double s_val[SP];
	__shared__ double s_b[SP];
	__shared__ int    s_col[SP];

	int i, index, local_cnt;
	double temp;

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		local_cnt = a.cnt[id];

		temp = 0.0;

		index = id;
		for(i=0; i<local_cnt; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_b[tid] = b[s_col[tid]];

			temp += s_val[tid] * s_b[tid];
		}

		index = id + a.row;
		c[index] = temp;
	}
}


void CUDA_TransitionMatrix(icrss &da, double *db, double *dc)
{
	int i, size;

	size = da.size();

	for(i=0; i<size; i++){
		TransitionMatrixTask<<<da[i].n / SP + 1 , SP>>>(da[i], db, dc);
	}

	return;
}


/*-----------------------------------------------------------------*/
__global__
void ReTransitionMatrixTask(icrs a, double *b, double *c)
{
	__shared__ double s_val[SP];
	__shared__ double s_b[SP];
	__shared__ int    s_col[SP];

	int i, index, local_cnt;
	double temp;

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		local_cnt = a.cnt[id];

		temp = 0.0;

		index = id;
		for(i=0; i<local_cnt; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_b[tid] = b[s_col[tid]];

			temp += s_val[tid] * s_b[tid];
		}

		index = id + a.row;
		c[index] += temp;
	}
}


void CUDA_ReTransitionMatrix(icrss &da, double *db, double *dc)
{
	int i, size;

	size = da.size();

	for(i=0; i<size; i++){
		ReTransitionMatrixTask<<<da[i].n / SP + 1 , SP>>>(da[i], db, dc);
	}

	return;
}

/*-----------------------------------------------------------------*/
__global__
void  TwoNormRelativeErrorTask(lu_icrs a, double *x, double *b, double *norm)
{
	__shared__ double s_norm[SP];
	__shared__ double s_val[SP];
	__shared__ double s_x[SP];
	__shared__ int    s_col[SP];

	int i, index, local_w;
	double temp, local_d, local_b, local_x;

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	s_norm[tid] = 0.0;

	if(id < a.n){
		index = id + a.row;
		local_w = a.cnt[id];
		local_d = a.d[id];
		local_x = x[index];
		local_b = b[index];

		temp = 0.0;

		index = id;
		for(i=0; i<local_w; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_x[tid] = x[s_col[tid]];

			temp -= s_val[tid] * s_x[tid];
		}
		temp -= local_d * local_x;
		temp += local_b;

		s_norm[tid] = temp * temp;
	}
	__syncthreads();

	SharedReduction<SP>(s_norm, tid);

	if(!tid){
		norm[blockIdx.x] += s_norm[0];
	}
}


double CUDA_TwoNormRelativeError(lu_icrss &da, double *dx, double *db)
{
	int i, size, temp, max;
	double norm;
	double *dsum, *sum;

	size = da.size();

	max = 0;
	for(i=0; i<size; i++){
		temp = da[i].n / SP + 1;
		max = (temp > max) ? temp : max;
	}

	sum = (double *)CUDA_MALLOC_HOST(sizeof(double) * max, "2 norm sum");
	CUDA_MALLOC((void**)&dsum, sizeof(double) * max, "2 norm malloc");

	for(i=0; i<max; i++){
		sum[i] = 0.0;
	}

	for(i=0; i<size; i++){
		TwoNormRelativeErrorTask<<< da[i].n / SP + 1, SP>>>
		(da[i], dx, db, dsum);
	}
	CUDA_LOAD(sum, dsum, sizeof(double) * max);

	norm = 0.0;
	for(i=0; i<max; i++){
		norm += sum[i];
	}
	CUDA_FREE_HOST(sum);
	CUDA_FREE(dsum);

	return norm;
}

/*-----------------------------------------------------------------*/

/*  */
__global__
void DotTask(double *A, double *B, int N, double *V, int width)
{
	__shared__ double s_V[SP];

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	int index = id;

	/* ɤν */
	s_V[tid] = 0.0;

	/* ʬ */
	while(index < N){
		s_V[tid] += A[index] * B[index];
		index += width;
	}
	__syncthreads();


	/* ɥ */
	SharedReduction<SP>(s_V, tid);


	/* Х¤ư */
	if(tid == 0)
		V[blockIdx.x] = s_V[0];
}


double CUDA_Dot(double *da, double *db, int np)
{
	int i;
	double sum, dot[SM];
	double *ddot;

	CUDA_MALLOC((void**)&ddot, sizeof(double) * SM, "dot malloc");

	DotTask<<< SM, SP>>>(da, db, np, ddot, SM*SP);

	CUDA_LOAD(dot, ddot, sizeof(double) * SM);

	sum = 0.0;
	for(i=0; i<SM; i++){
		sum += dot[i];
	}

	CUDA_FREE(ddot);
	return sum;
}

/*-----------------------------------------------------------------*/

bool CUDA_TerminalVal(object *obj)
{
	level_object *lv_obj;
	double norm, residual;

	lv_obj = &obj->level[0];

	norm = CUDA_TwoNormRelativeError(lv_obj->da, lv_obj->dx, lv_obj->db);

	residual = sqrt(norm / obj->first_residual);

	printf("residual:%e\n", residual);

	if(residual < obj->residual){
		return true;
	}else{
		return false;
	}
}

/*-----------------------------------------------------------------*/

__global__ void VectorLessTask(double *x, double *y, int n, bool *v)
{
	__shared__ bool sv[SP];

	/* ᥤid,åidγ*/
	const int tid = threadIdx.x;
	const int id = blockIdx.x * blockDim.x + threadIdx.x;
	int index = id;

	/* ɤν */
	sv[tid] = true;

	/* ʬ */
	while(index < n){
		sv[tid] = sv[tid] && (fabs(x[index]) < y[index]);
		index += SM*SP;
	}
	__syncthreads();


	/* ɥ */
	SharedReduction<SP>(sv, tid);


	/* ХӤư */
	if(tid == 0)
		v[blockIdx.x] = sv[0];
}

bool CUDA_VectorLess(double *dx, double *dy, int n)
{
	int i;
	bool sum, dot[SM];
	bool *ddot;

	CUDA_MALLOC((void**)&ddot, sizeof(bool) * SM, "vector less malloc");

	VectorLessTask<<< SM, SP>>>(dx, dy, n, ddot);

	CUDA_LOAD(dot, ddot, sizeof(bool) * SM);

	sum = true;
	for(i=0; i<SM; i++){
		sum = sum && dot[i];
	}

	CUDA_FREE(ddot);

	return sum;
}

bool CUDA_TerminalVec(object *obj)
{
	bool flag;
	level_object *lv_obj;
	lv_obj = &obj->level[0];

	CUDA_ResidualError(lv_obj->da, lv_obj->dx, lv_obj->db, lv_obj->de);
	flag = CUDA_VectorLess(lv_obj->de, obj->threshold, lv_obj->np);

	printf("\n");

	return flag;
}

/*-----------------------------------------------------------------*/

__global__ void FUNC_0(double *p, double *r, double *v, double beta, double omega, int n)
{
	const int id = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < n){
		p[id] = r[id] + beta * (p[id] - omega * v[id]);
	}
}

void CUDA_FUNC_0(double *dp, double *dr, double *dv, double beta, double omega, int np)
{
	FUNC_0<<< np / SP + 1, SP >>>(dp, dr, dv, beta, omega, np);
	return;
}

/*-----------------------------------------------------------------*/

__global__ void FUNC_1(double *s, double *r, double *v, double alpha, int n)
{
	const int id = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < n){
		s[id] = r[id] - alpha * v[id];
	}
}

void CUDA_FUNC_1(double *ds, double *dr, double *dv, double alpha, int np)
{
	FUNC_1<<< np / SP + 1, SP >>>(ds, dr, dv, alpha, np);
	return;
}

/*-----------------------------------------------------------------*/

__global__ void FUNC_2(double *x, double *pt, double *st, double alpha, double omega, int n)
{
	const int id = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < n){
		x[id] += alpha * pt[id] + omega * st[id];
	}
}

void CUDA_FUNC_2(double *dx, double *dpt, double *dst, double alpha, double omega, int np)
{
	FUNC_2<<< np / SP + 1, SP >>>(dx, dpt, dst, alpha, omega, np);
	return;
}

/*-----------------------------------------------------------------*/

__global__
void MMULT(lu_icrs a, double *b, double *c)
{
	__shared__ double s_val[SP];
	__shared__ double s_b[SP];
	__shared__ int    s_col[SP];

	int i, index, local_w;
	double temp, local_d, local_b;

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		index = id + a.row;
		local_w = a.cnt[id];
		local_d = a.d[id];
		local_b = b[index];

		temp = 0.0;

		index = id;
		for(i=0; i<local_w; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_b[tid] = b[s_col[tid]];

			temp += s_val[tid] * s_b[tid];
		}
		temp += local_d * local_b;

		index = id + a.row;
		c[index] = temp;
	}
}


void CUDA_MMULT(lu_icrss &da, double *db, double *dc)
{
	int i, size;

	size = da.size();

	for(i=0; i<size; i++){
		MMULT<<< da[i].n / SP + 1, SP>>>(da[i], db, dc);
	}
	return;
}
